@@ -20,6 +20,7 @@ import (
 	"runtime"
 	"strings"
 
+	gardenercontrolplane "github.com/danielfoehrkn/kubeswitch/pkg/subcommands/gardener"
 	"github.com/danielfoehrkn/kubeswitch/pkg/subcommands/history"
 	"github.com/danielfoehrkn/kubeswitch/pkg/subcommands/ns"
 	"github.com/sirupsen/logrus"
@@ -184,6 +185,10 @@ func init() {
 		Short:   "Change the current namespace",
 		Long:    `Search namespaces in the current cluster and change to it.`,
 		RunE: func(cmd *cobra.Command, args []string) error {
+			if len(args) == 1 && len(args[0]) > 0 {
+				return ns.SwitchToNamespace(args[0], getKubeconfigPathFromFlag())
+			}
+
 			return ns.SwitchNamespace(getKubeconfigPathFromFlag(), stateDirectory, noIndex)
 		},
 	}
@@ -203,6 +208,28 @@ func init() {
 		},
 	}
 
+	gardenerCmd := &cobra.Command{
+		Use:   "gardener",
+		Short: "gardener specific commands",
+		Long:  `Commands that can only be used if a Gardener store is configured.`,
+	}
+
+	controlplaneCmd := &cobra.Command{
+		Use:   "controlplane",
+		Short: "Switch to the Shoot's controlplane",
+		RunE: func(cmd *cobra.Command, args []string) error {
+			stores, _, err := initialize()
+			if err != nil {
+				return err
+			}
+
+			_, err = gardenercontrolplane.SwitchToControlplane(stores, getKubeconfigPathFromFlag())
+			return err
+		},
+	}
+
+	gardenerCmd.AddCommand(controlplaneCmd)
+
 	listContextsCmd := &cobra.Command{
 		Use:     "list-contexts",
 		Short:   "List all available contexts without fuzzy search",
@@ -308,10 +335,12 @@ func init() {
 	rootCommand.AddCommand(lastContextCmd)
 	rootCommand.AddCommand(aliasContextCmd)
 	rootCommand.AddCommand(versionCmd)
+	rootCommand.AddCommand(gardenerCmd)
 
 	setContextCmd.SilenceUsage = true
 	aliasContextCmd.SilenceErrors = true
 	aliasRmCmd.SilenceErrors = true
+	namespaceCommand.SilenceErrors = true
 
 	setFlagsForContextCommands(setContextCmd)
 	setFlagsForContextCommands(listContextsCmd)
@@ -322,6 +351,7 @@ func init() {
 	setFlagsForContextCommands(aliasContextCmd)
 
 	setCommonFlags(namespaceCommand)
+	setControlplaneCommandFlags(controlplaneCmd)
 }
 
 func NewCommandStartSwitcher() *cobra.Command {
@@ -386,6 +416,15 @@ func setCommonFlags(command *cobra.Command) {
 		"path to the local directory used for storing internal state.")
 }
 
+func setControlplaneCommandFlags(cmd *cobra.Command) {
+	setCommonFlags(cmd)
+	cmd.Flags().StringVar(
+		&configPath,
+		"config-path",
+		os.ExpandEnv("$HOME/.kube/switch-config.yaml"),
+		"path on the local filesystem to the configuration file.")
+}
+
 func initialize() ([]store.KubeconfigStore, *types.Config, error) {
 	if showDebugLogs {
 		logrus.SetLevel(logrus.DebugLevel)
