@@ -51,12 +51,13 @@ func SetContext(desiredContext string, stores []store.KubeconfigStore, config *t
 
 		kubeconfigStore := *discoveredContext.Store
 
-		contextWithoutPrefix := discoveredContext.Path
+		contextWithoutPrefix := discoveredContext.Name
 		if len(kubeconfigStore.GetContextPrefix(discoveredContext.Path)) > 0 && strings.HasPrefix(discoveredContext.Name, kubeconfigStore.GetContextPrefix(discoveredContext.Path)) {
 			contextWithoutPrefix = strings.TrimPrefix(discoveredContext.Name, fmt.Sprintf("%s/", kubeconfigStore.GetContextPrefix(discoveredContext.Path)))
 		}
 
-		if desiredContext == discoveredContext.Name || desiredContext == contextWithoutPrefix || desiredContext == discoveredContext.Alias {
+		matchesContextWithoutPrefix := desiredContext == contextWithoutPrefix
+		if desiredContext == discoveredContext.Name || matchesContextWithoutPrefix || desiredContext == discoveredContext.Alias {
 			kubeconfigData, err := kubeconfigStore.GetKubeconfigForPath(discoveredContext.Path)
 			if err != nil {
 				return nil, err
